#include <stdio.h>
#include <string.h>
#include <termios.h>
//
//  $BC<Kv@_Dj(B
//
int cbreak(int fd)
{
  struct termios buf;
  if (tcgetattr(fd, &buf) < 0) return -1;
  buf.c_lflag &= ~(ECHO | ICANON);
  buf.c_cc[VMIN] = 1;
  buf.c_cc[VTIME] = 0;
  if (tcsetattr(fd, TCSAFLUSH, &buf) < 0) return -1;
  return 0;
}
//
//  VT100$B%+!<%=%k@)8f(B
//
#define reverse()  printf("\033[7m")
#define normal()  printf("\033[m")
#define bol()    printf("\r")
#define erase()    printf("\033[K")
#define fullerase()  printf("\r\033[K")
#define up()    printf("\033[A")
#define right(i)  if((i)>0) printf("\033[%dC",i)
//
//  $BC18l<-=q(B/$BNcJ8<-=q(B
//
#define WORDDIC "word.pp"
#define MAXWORDS 40000
unsigned char *word[MAXWORDS],*yomi[MAXWORDS];
int nwords = 0;
#define PHRASEDIC "phrase.pp"
#define MAXPHRASES 50000
unsigned char *pcontext[MAXPHRASES],*pword[MAXPHRASES],*pyomi[MAXPHRASES];
int nphrases = 0;

#define MAXLINELEN 1000
#define MAXWLEN 100

void readdic()
{
  FILE *f;
  char linebuf[MAXLINELEN];
  char context[MAXWLEN],y[MAXWLEN],w[MAXWLEN];

  if((f = fopen(WORDDIC,"r")) == NULL) exit(0);
  while(fgets(linebuf,MAXLINELEN,f) && nwords < MAXWORDS){
    sscanf(linebuf,"%s %s",y,w);
    word[nwords] = strdup(w);
    yomi[nwords] = strdup(y);
    nwords++;
  }
  fclose(f);

  if((f = fopen(PHRASEDIC,"r")) == NULL) exit(0);
  while(fgets(linebuf,MAXLINELEN,f) && nphrases < MAXPHRASES){
    sscanf(linebuf,"%s %s %s",context,y,w);
    pcontext[nphrases] = strdup(context);
    pword[nphrases] = strdup(w);
    pyomi[nphrases] = strdup(y);
    nphrases++;
  }
  fclose(f);
}
//
//  $B8uJd(B
//
#define MAXCANDS 50
unsigned char cands[MAXCANDS][MAXWLEN];
int ncands = 0;
int curcand = -1;  // $B8=:_A*Br$5$l$F$$$k8uJd$N%$%s%G%/%9(B

char outbuf[MAXLINELEN] = "";  // $B=PNOJ8;zNs(B
int outbuflen = 0;
//
//  $B%Q%?%s(B/$B8!:w(B
//
char pat[100] = "";
int exact = 0;

void resetpat()
{
  strcpy(pat,"");
  exact = 0;
  curcand = -1;
}
void search()
{
  int i,j;
  int plen,patlen;
  outbuflen = strlen(outbuf);
  patlen = strlen(pat);
  ncands = 0;
  for(i=0;i<nphrases && ncands < MAXCANDS-1;i++){
    plen = strlen(pcontext[i]);
    if(outbuflen >= plen &&
       strncmp(pcontext[i],outbuf+outbuflen-plen,plen) == 0 &&
       (exact ? strcmp(pat,pyomi[i]) : strncmp(pat,pyomi[i],patlen)) == 0){
      for(j=0;j<ncands;j++)
        if(strcmp(pword[i],cands[j])==0) break;
      if(j < ncands) continue;
      strcpy(cands[ncands++],pword[i]);
    }
  }
  for(i=0;i<nwords && ncands < MAXCANDS-1;i++){
    if((exact ? strcmp(pat,yomi[i]) : strncmp(pat,yomi[i],patlen)) == 0){
      for(j=0;j<ncands;j++)
        if(strcmp(word[i],cands[j])==0) break;
      if(j < ncands) continue;
      strcpy(cands[ncands++],word[i]);
    }
  }
}
//
//  $B3NDjJ8;zNs(B/$B8uJd$NI=<((B
//
void display()
{
  int i,col = strlen(outbuf);

  printf("\n");
  fullerase();
  right(col);
  for(i=curcand+1;i<ncands && i<=curcand+5;i++){
    if((col += (strlen(cands[i])+1)) > 78) break;
    printf("%s ",cands[i]);
  }
  up();
  fullerase();
  printf("%s",outbuf);
  if(curcand >= 0){
    reverse();
    printf("%s",cands[curcand]);
    normal();
  }
  fflush(stdout);
}
//
//  $B%a%$%s%k!<%A%s(B
//
int main(int argc, char **argv)
{
  char s[2] = " ";

  cbreak(0);
  readdic();
  search();
  display();
  while(read(0,s,1)){
    switch(*s){
    case 'c': // $B%Q%?%s%/%j%"(B
      resetpat();
      search();
      break;
    case 'z': // $B3NDj(B/$B2~9T(B
      strcat(outbuf,cands[curcand]);
      resetpat();
      display();
      printf("\n");

      strcpy(outbuf,"");
      search();
      break;
    case 'v': // $B40A40lCW8!:w;XDj(B
      if(! exact){
        exact = 1;
        search();
        curcand = 0;
      }
      break;
    case 'x': // $BA08uJd(B/$B8eB`(B
      if(curcand >= 0){ // $BA08uJd(B
        curcand--;
      }
      else { // $B8eB`(B
        outbuflen = strlen(outbuf);
        if(outbuflen > 1) outbuf[outbuflen-2]='\0';
        resetpat();
        search();
      }
      break;
    case ' ': // $B<!8uJd(B
      curcand++;
      break;
    case 'a': case 's': case 'd': case 'f':
      if(curcand>=0 && strlen(pat) > 0){ // $BA0$N8uJd$r3NDj(B
        strcat(outbuf,cands[curcand]);
        resetpat();
      }
      strcat(pat,s);
      search();
      break;
    }
    display();
  }
}

